import os
import logging
import shutil
from tempfile import mkstemp


def generate(spec):
    '''
    resolves audiospec-static

    Recognized argument is  "paths" (list of static paths)
    '''
    if 'paths' not in spec:
        raise ValueError("Malformed audiospec: missing 'paths'")

    for path in spec['paths']:
        if not os.path.exists(path):
            logging.warn("Can't find requested path: %s" % path)
            continue
        tmp = mkstemp(suffix=os.path.splitext(path)[-1],
                      prefix='audiogen-static-')
        os.close(tmp[0])
        shutil.copy(path, tmp[1])
        yield 'file://{}'.format(tmp[1])
generate.description = 'Picks always the same specified file'



from datetime import datetime

from pytimeparse.timeparse import timeparse
from flask_wtf import Form
from wtforms import StringField, DateTimeField, validators, \
        SubmitField, ValidationError


class FrequencyAlarmForm(Form):
    nick = StringField('Alarm nick', validators=[validators.required()],
                       description='A simple name to recognize this alarm')
    interval = StringField('Frequency',
                           validators=[validators.required()],
                           description='in seconds, or human-readable '
                           '(like 9w3d12h)')
    start = DateTimeField('Start date and time',
                          validators=[validators.optional()],
                          description='Date before which no alarm will ring, '
                          'expressed as YYYY-MM-DD HH:MM:SS; if omitted, the '
                          'alarm will always ring')
    end = DateTimeField('End date and time',
                        validators=[validators.optional()],
                        description='Date after which no alarm will ring, '
                        'expressed as YYYY-MM-DD HH:MM:SS; if omitted, the '
                        'alarm will always ring')
    submit = SubmitField('Submit')

    def populate_from_timespec(self, timespec):
        if 'nick' in timespec:
            self.nick.data = timespec['nick']
        if 'start' in timespec:
            self.start.data = datetime.fromtimestamp(timespec['start'])
        if 'end' in timespec:
            self.end.data = datetime.fromtimestamp(timespec['end'])
        self.interval.data = timespec['interval']

    def validate_interval(form, field):
        try:
            int(field.data)
        except ValueError:
            if timeparse(field.data) is None:
                raise ValidationError("interval must either be a number "
                                      "(in seconds) or a human-readable "
                                      "string like '1h2m'  or '1d12h'")


def frequencyalarm_receive(form):
    obj = {
        'kind': 'frequency',
        'nick': form.nick.data,
        'interval': form.interval.data,
    }
    if form.start.data:
        obj['start'] = int(form.end.data.strftime('%s'))
    else:
        obj['start'] = 0
    if form.end.data:
        obj['end'] = int(form.end.data.strftime('%s'))
    return obj



from __future__ import print_function
import sys
import argparse
from .entrypoints_utils import get_one_entrypoint
import json
from logging import getLogger
log = getLogger('audiogen')


def get_audiogenerator(kind):
    '''Messes with entrypoints to return an audiogenerator function'''
    return get_one_entrypoint('larigira.audiogenerators', kind)


def get_parser():
    parser = argparse.ArgumentParser(
        description='Generate audio and output paths')
    parser.add_argument('audiospec', metavar='AUDIOSPEC', type=str, nargs=1,
                        help='filename for audiospec, formatted in json')
    return parser


def read_spec(fname):
    try:
        if fname == '-':
            return json.load(sys.stdin)
        with open(fname) as buf:
            return json.load(buf)
    except ValueError:
        sys.stderr.write("Error: invalid JSON\n")
        sys.exit(1)


def check_spec(spec):
    if 'kind' not in spec:
        yield "Missing field 'kind'"


def audiogenerate(spec):
    gen = get_audiogenerator(spec['kind'])
    return tuple(gen(spec))


def main():
    '''Main function for the "larigira-audiogen" executable'''
    args = get_parser().parse_args()
    spec = read_spec(args.audiospec[0])
    errors = tuple(check_spec(spec))
    if errors:
        log.error("Errors in audiospec")
        for err in errors:
            sys.stderr.write('Error: {}\n'.format(err))
        sys.exit(1)
    for path in audiogenerate(spec):
        print(path)


if __name__ == '__main__':
    main()



from logging import getLogger
log = getLogger('entrypoints_utils')
from pkg_resources import iter_entry_points


def get_one_entrypoint(group, kind):
    '''Messes with entrypoints to return an entrypoint of a given group/kind'''
    points = tuple(iter_entry_points(group=group, name=kind))
    if not points:
        raise ValueError('cant find an entrypoint %s:%s' % (group, kind))
    if len(points) > 1:
        log.warning("Found more than one timeform for %s:%s" % (group, kind))
    return points[0].load()


def get_avail_entrypoints(group):
    return [e.name for e in iter_entry_points(group=group)]



from datetime import datetime

from flask_wtf import Form
from wtforms import StringField, DateTimeField, validators, SubmitField


class SingleAlarmForm(Form):
    nick = StringField('Alarm nick', validators=[validators.required()],
                       description='A simple name to recognize this alarm')
    dt = DateTimeField('Date and time', validators=[validators.required()],
                       description='Date to ring on, expressed as '
                       'YYYY-MM-DD HH:MM:SS')
    submit = SubmitField('Submit')

    def populate_from_timespec(self, timespec):
        if 'nick' in timespec:
            self.nick.data = timespec['nick']
        if 'timestamp' in timespec:
            self.dt.data = datetime.fromtimestamp(timespec['timestamp'])


def singlealarm_receive(form):
    return {
        'kind': 'single',
        'nick': form.nick.data,
        'timestamp': int(form.dt.data.strftime('%s'))
    }



import flask_wtf
from wtforms import StringField, validators, SubmitField, IntegerField


class Form(flask_wtf.Form):
    nick = StringField('Audio nick', validators=[validators.required()],
                       description='A simple name to recognize this audio')
    path = StringField('Path', validators=[validators.required()],
                       description='Full path to source directory')
    howmany = IntegerField('Number', validators=[validators.optional()],
                           default=1,
                           description='How many songs to be picked'
                           'from this dir; defaults to 1')
    submit = SubmitField('Submit')

    def populate_from_audiospec(self, audiospec):
        if 'nick' in audiospec:
            self.nick.data = audiospec['nick']
        if 'paths' in audiospec:
            self.path.data = audiospec['paths'][0]
        if 'howmany' in audiospec:
            self.howmany.data = audiospec['howmany']
        else:
            self.howmany.data = 1


def receive(form):
    return {
        'kind': 'randomdir',
        'nick': form.nick.data,
        'paths': [form.path.data],
        'howmany': form.howmany.data or 1
    }




from flask_wtf import Form
from wtforms import StringField, validators, SubmitField, ValidationError


class ScriptAudioForm(Form):
    nick = StringField('Audio nick', validators=[validators.required()],
                       description='A simple name to recognize this audio')
    name = StringField('Name', validators=[validators.required()],
                       description='filename (NOT path) of the script')
    args = StringField('Arguments',
                       description='arguments, separated by spaces')
    submit = SubmitField('Submit')

    def populate_from_audiospec(self, audiospec):
        if 'nick' in audiospec:
            self.nick.data = audiospec['nick']
        if 'name' in audiospec:
            self.name.data = audiospec['name']
        if 'args' in audiospec:
            self.args.data = audiospec['args']

    def validate_name(form, field):
        if '/' in field.data:
            raise ValidationError("Name cannot have slashes: "
                                  "it's a name, not a path")


def scriptaudio_receive(form):
    return {
        'kind': 'script',
        'nick': form.nick.data,
        'name': form.name.data,
        'args': form.args.data
    }



'''
This component will look for files to be removed. There are some assumptions:
    * Only files in $TMPDIR are removed. Please remember that larigira has its
      own specific TMPDIR
    * MPD URIs are parsed, and only file:/// is supported
'''
import os
import logging
import mpd


class UnusedCleaner:
    def __init__(self, conf):
        self.conf = conf
        self.waiting_removal_files = set()
        self.log = logging.getLogger(self.__class__.__name__)

    def _get_mpd(self):
        mpd_client = mpd.MPDClient(use_unicode=True)
        mpd_client.connect(self.conf['MPD_HOST'], self.conf['MPD_PORT'])
        return mpd_client

    def watch(self, uri):
        '''
        adds fpath to the list of "watched" file

        as soon as it leaves the mpc playlist, it is removed
        '''
        if not uri.startswith('file:///'):
            raise ValueError('not a file URI')
        fpath = uri[len('file://'):]
        if not os.path.exists(fpath):
            self.log.warning('a path that does not exist is being monitored')
        self.waiting_removal_files.add(fpath)

    def check_playlist(self):
        '''check playlist + internal watchlist to see what can be removed'''
        mpd = self._get_mpd()
        files_in_playlist = {song['file'] for song in mpd.playlistid()
                             if song['file'].startswith('/')}
        for fpath in self.waiting_removal_files - files_in_playlist:
            # we can remove it!
            self.log.debug('removing unused: {}'.format(fpath))
            self.waiting_removal_files.remove(fpath)
            if os.path.exists(fpath):
                os.unlink(fpath)



import gevent


class ParentedLet(gevent.Greenlet):
    '''
    ParentedLet is just a helper subclass that will help you when your
    greenlet main duty is to "signal" things to a parent_queue.

    It won't save you much code, but "standardize" messages and make explicit
    the role of that greenlet
    '''
    def __init__(self, queue):
        gevent.Greenlet.__init__(self)
        self.parent_queue = queue
        self.parent_greenlet = None
        self.tracker = None  # set this to recognize easily

    def parent_msg(self, kind, *args):
        return {
            'emitter': self,
            'class': self.__class__.__name__,
            'tracker': self.tracker,
            'kind': kind,
            'args': args
        }

    def send_to_parent(self, kind, *args):
        self.parent_queue.put(self.parent_msg(kind, *args))

    def _run(self):
        if not hasattr(self, 'do_business'):
            raise Exception("do_business method not implemented by %s" %
                            self.__class__.__name__)
        for msg in self.do_business():
            self.send_to_parent(*msg)


class Timer(ParentedLet):
    '''continously sleeps some time, then send a "timer" message to parent'''
    def __init__(self, milliseconds, queue):
        ParentedLet.__init__(self, queue)
        self.ms = milliseconds

    def parent_msg(self, kind, *args):
        msg = ParentedLet.parent_msg(self, kind, *args)
        msg['period'] = self.ms
        return msg

    def do_business(self):
        while True:
            gevent.sleep(self.ms / 1000.0)
            yield ('timer', )



from __future__ import print_function
import argparse
import json

from .event import EventModel
from .config import get_conf


def main_list(args):
    m = EventModel(args.file)
    for alarm, action in m.get_all_alarms_expanded():
        print(json.dumps(dict(alarm=alarm, action=action), indent=4))


def main_add(args):
    m = EventModel(args.file)
    m.add_event(dict(kind='frequency', interval=args.interval, start=1),
                [dict(kind='mpd', howmany=1)]
                )


def main():
    conf = get_conf()
    p = argparse.ArgumentParser()
    p.add_argument('-f', '--file', help="Filepath for DB", required=False,
                   default=conf['DB_URI'])
    sub = p.add_subparsers()
    sub_list = sub.add_parser('list')
    sub_list.set_defaults(func=main_list)
    sub_add = sub.add_parser('add')
    sub_add.add_argument('--interval', type=int, default=3600)
    sub_add.set_defaults(func=main_add)

    args = p.parse_args()
    args.func(args)

if __name__ == '__main__':
    main()



from __future__ import print_function
import logging
log = logging.getLogger('time-every')
from datetime import datetime, timedelta

from pytimeparse.timeparse import timeparse


def getdate(val):
    if type(val) is int:
        return datetime.fromtimestamp(val)
    return val


class Alarm(object):
    def __init__(self):
        pass

    def next_ring(self, current_time=None):
        '''if current_time is None, it is now()

        returns the next time it will ring; or None if it will not anymore
        '''
        raise NotImplementedError()

    def has_ring(self, time=None):
        raise NotImplementedError()

    def all_rings(self, current_time=None):
        '''
        all future rings
        this, of course, is an iterator (they could be infinite)
        '''
        ring = self.next_ring(current_time)
        while ring is not None:
            yield ring
            ring = self.next_ring(ring)


class SingleAlarm(Alarm):
    '''
    rings a single time
    '''
    description = 'Only once, at a specified date and time'

    def __init__(self, obj):
        self.dt = getdate(obj['timestamp'])

    def next_ring(self, current_time=None):
        '''if current_time is None, it is now()'''
        if current_time is None:
            current_time = datetime.now()
        if current_time >= self.dt:
            return None
        return self.dt

    def has_ring(self, current_time=None):
        if current_time is None:
            current_time = datetime.now()
        return current_time == self.dt


class FrequencyAlarm(Alarm):
    '''
    rings on {t | exists a k integer >= 0 s.t. t = start+k*t, start<t<end}
    '''
    description = 'Events at a specified frequency. Example: every 30minutes'

    def __init__(self, obj):
        self.start = getdate(obj['start'])
        try:
            self.interval = int(obj['interval'])
        except ValueError:
            self.interval = timeparse(obj['interval'])
        assert type(self.interval) is int
        self.end = getdate(obj['end']) if 'end' in obj else None

    def next_ring(self, current_time=None):
        '''if current_time is None, it is now()'''
        if current_time is None:
            current_time = datetime.now()
        if self.end is not None and current_time > self.end:
            return None
        if current_time < self.start:
            return self.start
        if self.end is not None:
            assert self.start <= current_time <= self.end
        else:
            assert self.start <= current_time
        n_interval = (
            (current_time - self.start).total_seconds() // self.interval
            ) + 1
        ring = self.start + timedelta(seconds=self.interval * n_interval)
        if ring == current_time:
            ring += timedelta(seconds=self.interval)
        if self.end is not None and ring > self.end:
            return None
        return ring

    def has_ring(self, current_time=None):
        if current_time is None:
            current_time = datetime.now()
        if not self.start >= current_time >= self.end:
            return False

        n_interval = (current_time - self.start).total_seconds() // \
            self.interval
        expected_time = self.start + \
            timedelta(seconds=self.interval * n_interval)
        return expected_time == current_time

    def __str__(self):
        return 'FrequencyAlarm(every %ds)' % self.interval



#!/usr/bin/env python
'''
Taken from flask-appconfig
'''

import json
import os

from xdg import BaseDirectory


def get_conf(prefix='LARIGIRA_'):
    '''This is where everyone should get configuration from'''
    conf_dir = BaseDirectory.save_config_path('larigira')
    conf = {}
    conf['CONTINOUS_AUDIODESC'] = dict(kind='mpd', howmany=1)
    conf['MPD_HOST'] = os.getenv('MPD_HOST', 'localhost')
    conf['MPD_PORT'] = int(os.getenv('MPD_PORT', '6600'))
    conf['CACHING_TIME'] = 10
    conf['DB_URI'] = os.path.join(conf_dir, 'db.json')
    conf['SCRIPTS_PATH'] = os.path.join(conf_dir, 'scripts')
    conf['BOOTSTRAP_SERVE_LOCAL'] = True
    conf['SECRET_KEY'] = 'Please replace me!'
    conf['MPD_WAIT_START'] = True
    conf['MPD_WAIT_START_RETRYSECS'] = 5
    conf['CHECK_SECS'] = 20   # period for checking playlist length
    conf['EVENT_TICK_SECS'] = 30  # period for scheduling events
    conf['DEBUG'] = False
    conf.update(from_envvars(prefix=prefix))
    return conf


def from_envvars(prefix=None, envvars=None, as_json=True):
    """Load environment variables in a dictionary

    Values are parsed as JSON. If parsing fails with a ValueError,
    values are instead used as verbatim strings.

    :param prefix: If ``None`` is passed as envvars, all variables from
                   ``environ`` starting with this prefix are imported. The
                   prefix is stripped upon import.
    :param envvars: A dictionary of mappings of environment-variable-names
                    to Flask configuration names. If a list is passed
                    instead, names are mapped 1:1. If ``None``, see prefix
                    argument.
    :param as_json: If False, values will not be parsed as JSON first.
    """
    conf = {}
    if prefix is None and envvars is None:
        raise RuntimeError('Must either give prefix or envvars argument')

    # if it's a list, convert to dict
    if isinstance(envvars, list):
        envvars = {k: None for k in envvars}

    if not envvars:
        envvars = {k: k[len(prefix):] for k in os.environ.keys()
                   if k.startswith(prefix)}

    for env_name, name in envvars.items():
        if name is None:
            name = env_name

        if env_name not in os.environ:
            continue

        if as_json:
            try:
                conf[name] = json.loads(os.environ[env_name])
            except ValueError:
                conf[name] = os.environ[env_name]
        else:
            conf[name] = os.environ[env_name]

    return conf



'''
main module to read and get informations about alarms
'''
from __future__ import print_function
import sys
from datetime import datetime
import argparse
import json
from .entrypoints_utils import get_one_entrypoint
from logging import getLogger
log = getLogger('timegen')


def get_timegenerator(kind):
    '''Messes with entrypoints to return an timegenerator function'''
    return get_one_entrypoint('larigira.timegenerators', kind)


def get_parser():
    parser = argparse.ArgumentParser(
        description='Generate "ring times" from a timespec')
    parser.add_argument('timespec', metavar='TIMESPEC', type=str, nargs=1,
                        help='filename for timespec, formatted in json')
    parser.add_argument('--now', metavar='NOW', type=int, nargs=1,
                        default=None,
                        help='Set a different "time", in unix epoch')
    parser.add_argument('--howmany', metavar='N', type=int, nargs=1,
                        default=[1],
                        help='Set a different "time", in unix epoch')
    return parser


def read_spec(fname):
    try:
        if fname == '-':
            return json.load(sys.stdin)
        with open(fname) as buf:
            return json.load(buf)
    except ValueError:
        sys.stderr.write("Error: invalid JSON\n")
        sys.exit(1)


def check_spec(spec):
    if 'kind' not in spec:
        yield "Missing field 'kind'"


def timegenerate(spec, now=None, howmany=1):
    Alarm = get_timegenerator(spec['kind'])
    generator = Alarm(spec)
    if now is not None:
        if type(now) is not datetime:
            now = datetime.fromtimestamp(now)
    for _ in range(howmany):
        now = generator.next_ring(current_time=now)
        yield now


def main():
    '''Main function for the "larigira-timegen" executable'''
    args = get_parser().parse_args()
    spec = read_spec(args.timespec[0])
    errors = tuple(check_spec(spec))
    if errors:
        log.error("Errors in timespec")
        for err in errors:
            sys.stderr.write('Error: {}\n'.format(err))
        sys.exit(1)
    now = None if args.now is None else args.now.pop()
    howmany = None if args.howmany is None else args.howmany.pop()
    for time in timegenerate(spec, now=now, howmany=howmany):
        print(time)



from __future__ import print_function
from gevent import monkey
monkey.patch_all(subprocess=True)
import logging
logging.getLogger('mpd').setLevel(logging.WARNING)
from datetime import datetime, timedelta

import gevent
from gevent.queue import Queue
from tinydb import TinyDB

from .eventutils import ParentedLet, Timer
from .timegen import timegenerate
from .audiogen import audiogenerate


class EventModel(object):
    def __init__(self, uri):
        self.uri = uri
        self.db = None
        self.reload()

    def reload(self):
        if self.db is not None:
            self.db.close()
        self.db = TinyDB(self.uri)
        self.actions = self.db.table('actions')
        self.alarms = self.db.table('alarms')

    def get_action_by_id(self, action_id):
        return self.actions.get(eid=action_id)

    def get_alarm_by_id(self, alarm_id):
        return self.alarms.get(eid=alarm_id)

    def get_actions_by_alarm(self, alarm):
        for action_id in alarm.get('actions', []):
            yield self.get_action_by_id(action_id)

    def get_all_alarms(self):
        return self.alarms.all()

    def get_all_actions(self):
        return self.actions.all()

    def get_all_alarms_expanded(self):
        for alarm in self.get_all_alarms():
            for action in self.get_actions_by_alarm(alarm):
                yield alarm, action

    def add_event(self, alarm, actions):
        action_ids = [self.add_action(a) for a in actions]
        alarm['actions'] = action_ids
        return self.alarms.insert(alarm)

    def add_action(self, action):
        return self.actions.insert(action)

    def add_alarm(self, alarm):
        return self.add_event(alarm, [])

    def update_alarm(self, alarmid, new_fields={}):
        return self.alarms.update(new_fields, eids=[alarmid])

    def update_action(self, actionid, new_fields={}):
        return self.actions.update(new_fields, eids=[actionid])


class Monitor(ParentedLet):
    '''
    Manages timegenerators and audiogenerators.
    
    The mechanism is partially based on ticks, partially on scheduled actions.
    Ticks are emitted periodically; at every tick, :func:`on_tick
    <larigira.event.Monitor.on_tick>` checks if any event is "near enough". If
    an event is near enough, it is ":func:`scheduled
    <larigira.event.Monitor.schedule>`": a greenlet is run which will wait for
    the right time, then generate the audio, then submit to Controller.

    The tick mechanism allows for events to be changed on disk: if everything
    was scheduled immediately, no further changes would be possible.
    The scheduling mechanism allows for more precision, catching exactly the
    right time. Being accurate only with ticks would have required very
    frequent ticks, which is cpu-intensive.
    '''
    def __init__(self, parent_queue, conf):
        ParentedLet.__init__(self, parent_queue)
        self.log = logging.getLogger(self.__class__.__name__)
        self.running = {}
        self.conf = conf
        self.q = Queue()
        self.model = EventModel(self.conf['DB_URI'])
        self.ticker = Timer(int(self.conf['EVENT_TICK_SECS']) * 1000, self.q)

    def _alarm_missing_time(self, timespec):
        now = datetime.now() + timedelta(seconds=self.conf['CACHING_TIME'])
        try:
            when = next(timegenerate(timespec, now=now))
        except:
            logging.exception("Could not generate "
                              "an alarm from timespec {}".format(timespec))
        if when is None:
            # expired
            return None
        delta = (when - now).total_seconds()
        assert delta > 0
        return delta

    def on_tick(self):
        '''
        this is called every EVENT_TICK_SECS.
        Checks every event in the DB (which might be slightly CPU-intensive, so
        it is advisable to run it in its own greenlet); if the event is "near
        enough", schedule it; if it is too far, or already expired, ignore it.
        '''
        self.model.reload()
        for alarm in self.model.get_all_alarms():
            actions = list(self.model.get_actions_by_alarm(alarm))
            if alarm.eid in self.running:
                continue
            delta = self._alarm_missing_time(alarm)
            # why this 2*EVENT_TICK_SECS? EVENT_TICK_SECS would be enough,
            # but it is "tricky"; any small delay would cause the event to be
            # missed
            if delta is not None and delta <= 2*self.conf['EVENT_TICK_SECS']:
                self.log.debug('Scheduling event {} ({}s) => {}'
                               .format(alarm.get('nick', alarm.eid),
                                       delta,
                                       [a.get('nick', a.eid) for a in actions]
                                       ))
                self.schedule(alarm, actions, delta)
            else:
                self.log.debug('Skipping event {}, too far ({}s)'
                               .format(alarm.get('nick', alarm.eid),
                                       delta))

    def schedule(self, timespec, audiospecs, delta=None):
        '''
        prepare an event to be run at a specified time with the specified
        actions; the DB won't be read anymore after this call.

        This means that this call should not be done too early, or any update
        to the DB will be ignored.
        '''
        if delta is None:
            delta = self._alarm_missing_time(timespec)

        audiogen = gevent.spawn_later(delta,
                                      self.process_action,
                                      timespec, audiospecs)
        audiogen.parent_greenlet = self
        audiogen.doc = 'Will wait {} seconds, then generate audio "{}"'.format(
            delta,
            ','.join(aspec.get('nick', '') for aspec in audiospecs),
        )
        self.running[timespec.eid] = {
            'greenlet': audiogen,
            'running_time': datetime.now() + timedelta(seconds=delta),
            'timespec': timespec,
            'audiospecs': audiospecs
        }

    def process_action(self, timespec, audiospecs):
        '''Generate audio and submit it to Controller'''
        if timespec.eid in self.running:
            del self.running[timespec.eid]
        else:
            self.log.warn('Timespec {} completed but not in running '
                          'registry; this is most likely a bug'.
                          format(timespec.get('nick', timespec.eid)))
        uris = []
        for audiospec in audiospecs:
            try:
                uris.extend(audiogenerate(audiospec))
            except:
                self.log.error('audiogenerate for {} failed'
                               .format(audiospec))
        self.send_to_parent('uris_enqueue',
                            dict(uris=uris,
                                 timespec=timespec,
                                 audiospecs=audiospecs,
                                 aids=[a.eid for a in audiospecs]))

    def _run(self):
        self.ticker.start()
        gevent.spawn(self.on_tick)
        while True:
            value = self.q.get()
            kind = value['kind']
            if kind in ('forcetick', 'timer'):
                gevent.spawn(self.on_tick)
            else:
                self.log.warning("Unknown message: %s" % str(value))



from __future__ import print_function

from flask_wtf import Form
from wtforms import StringField, validators, SubmitField


class StaticAudioForm(Form):
    nick = StringField('Audio nick', validators=[validators.required()],
                       description='A simple name to recognize this audio')
    path = StringField('Path', validators=[validators.required()],
                       description='Full path to audio file')
    submit = SubmitField('Submit')

    def populate_from_audiospec(self, audiospec):
        if 'nick' in audiospec:
            self.nick.data = audiospec['nick']
        if 'paths' in audiospec:
            self.path.data = audiospec['paths'][0]


def staticaudio_receive(form):
    return {
        'kind': 'static',
        'nick': form.nick.data,
        'paths': [form.path.data]
    }



'''
This module is for the main application logic
'''
from __future__ import print_function
from gevent import monkey
monkey.patch_all(subprocess=True)

import sys
import os
import tempfile
import signal
from time import sleep
import logging

import gevent
from gevent.wsgi import WSGIServer

from .mpc import Controller, get_mpd_client
from .config import get_conf
from .rpc import create_app


def on_main_crash(*args, **kwargs):
    print('A crash occurred in "main" greenlet. Aborting...')
    sys.exit(1)


class Larigira(object):
    def __init__(self):
        self.log = logging.getLogger('larigira')
        self.conf = get_conf()
        self.controller = Controller(self.conf)
        self.controller.link_exception(on_main_crash)
        self.http_server = WSGIServer(('', 5000),
                                      create_app(self.controller.q, self))

    def start(self):
        self.controller.start()
        self.http_server.start()


def main():
    tempfile.tempdir = os.environ['TMPDIR'] = os.path.join(
        os.getenv('TMPDIR', '/tmp'),
        'larigira.%d' % os.getuid())
    if not os.path.isdir(os.environ['TMPDIR']):
        os.makedirs(os.environ['TMPDIR'])
    log_format = '%(asctime)s|%(levelname)s[%(name)s:%(lineno)d] %(message)s'
    logging.basicConfig(level=logging.DEBUG if get_conf()['DEBUG'] else logging.INFO,
                        format=log_format,
                        datefmt='%H:%M:%S')
    if(get_conf()['MPD_WAIT_START']):
        while True:
            try:
                get_mpd_client(get_conf())
            except Exception:
                logging.debug("Could not connect to MPD, waiting")
                sleep(int(get_conf()['MPD_WAIT_START_RETRYSECS']))
            else:
                logging.info("MPD ready!")
                break



    larigira = Larigira()
    larigira.start()

    def sig(*args):
        print('invoked sig', args)
        larigira.controller.q.put(dict(kind='signal', args=args))
    for signum in (signal.SIGHUP, signal.SIGALRM):
        gevent.signal(signum, sig, signum)
    gevent.wait()

if __name__ == '__main__':
    main()



from logging import getLogger
log = getLogger('timeform')
from .entrypoints_utils import get_one_entrypoint


def get_timeform(kind):
    '''Messes with entrypoints to return a TimeForm'''
    for group in ('larigira.timeform_create', 'larigira.timeform_receive'):
        yield get_one_entrypoint(group, kind)


def get_audioform(kind):
    '''Messes with entrypoints to return a AudioForm'''
    for group in ('larigira.audioform_create', 'larigira.audioform_receive'):
        yield get_one_entrypoint(group, kind)



'''
This module contains a flask blueprint for db administration stuff

Templates are self-contained in this directory
'''
from __future__ import print_function

from flask import current_app, Blueprint, render_template, jsonify, abort, \
    request, redirect, url_for

from larigira.entrypoints_utils import get_avail_entrypoints
from larigira.audiogen import get_audiogenerator
from larigira.timegen import get_timegenerator
from larigira import forms
db = Blueprint('db', __name__, url_prefix='/db', template_folder='templates')


def get_model():
    return current_app.larigira.controller.monitor.model


@db.route('/')
def home():
    return render_template('dbadmin_base.html')


@db.route('/list')
def list():
    model = current_app.larigira.controller.monitor.model
    alarms = tuple(model.get_all_alarms())
    events = [(alarm, model.get_actions_by_alarm(alarm))
              for alarm in alarms]
    return render_template('list.html', events=events)


@db.route('/add/time')
def addtime():
    kinds = get_avail_entrypoints('larigira.timeform_create')

    def gen_info(gen):
        return dict(description=getattr(gen, 'description', ''))
    info = {kind: gen_info(get_timegenerator(kind))
            for kind in kinds}
    return render_template('add_time.html', kinds=kinds, info=info)


@db.route('/edit/time/<int:alarmid>', methods=['GET', 'POST'])
def edit_time(alarmid):
    model = get_model()
    timespec = model.get_alarm_by_id(alarmid)
    kind = timespec['kind']
    Form, receiver = tuple(forms.get_timeform(kind))
    form = Form()
    if request.method == 'GET':
        form.populate_from_timespec(timespec)
    if request.method == 'POST' and form.validate():
        data = receiver(form)
        model.update_alarm(alarmid, data)
        model.reload()
        return redirect(url_for('db.list', _anchor='event-%d' % alarmid))
    return render_template('add_time_kind.html',
                           form=form,
                           kind=kind,
                           mode='edit',
                           )


@db.route('/add/time/<kind>', methods=['GET', 'POST'])
def addtime_kind(kind):
    Form, receiver = tuple(forms.get_timeform(kind))
    form = Form()
    if request.method == 'POST' and form.validate():
        data = receiver(form)
        eid = get_model().add_alarm(data)
        return redirect(url_for('db.edit_event', alarmid=eid))

    return render_template('add_time_kind.html',
                           form=form, kind=kind, mode='add')


@db.route('/add/audio')
def addaudio():
    kinds = get_avail_entrypoints('larigira.audioform_create')

    def gen_info(gen):
        return dict(description=getattr(gen, 'description', ''))
    info = {kind: gen_info(get_audiogenerator(kind))
            for kind in kinds}
    return render_template('add_audio.html', kinds=kinds, info=info)


@db.route('/add/audio/<kind>', methods=['GET', 'POST'])
def addaudio_kind(kind):
    Form, receiver = tuple(forms.get_audioform(kind))
    form = Form()
    if request.method == 'POST' and form.validate():
        data = receiver(form)
        model = current_app.larigira.controller.monitor.model
        eid = model.add_action(data)
        return jsonify(dict(inserted=eid, data=data))

    return render_template('add_audio_kind.html', form=form, kind=kind)


@db.route('/edit/audio/<int:actionid>', methods=['GET', 'POST'])
def edit_audio(actionid):
    model = get_model()
    audiospec = model.get_action_by_id(actionid)
    kind = audiospec['kind']
    Form, receiver = tuple(forms.get_audioform(kind))
    form = Form()
    if request.method == 'GET':
        form.populate_from_audiospec(audiospec)
    if request.method == 'POST' and form.validate():
        data = receiver(form)
        model.update_action(actionid, data)
        model.reload()
        return redirect(url_for('db.list'))
    return render_template('add_audio_kind.html',
                           form=form,
                           kind=kind,
                           mode='edit',
                           )


@db.route('/edit/event/<alarmid>')
def edit_event(alarmid):
    model = current_app.larigira.controller.monitor.model
    alarm = model.get_alarm_by_id(int(alarmid))
    if alarm is None:
        abort(404)
    allactions = model.get_all_actions()
    print('all', allactions)
    actions = tuple(model.get_actions_by_alarm(alarm))
    return render_template('edit_event.html',
                           alarm=alarm, all_actions=allactions,
                           actions=actions)


@db.route('/api/alarm/<alarmid>/actions', methods=['POST'])
def change_actions(alarmid):
    new_actions = request.form.getlist('actions[]')
    if new_actions is None:
        new_actions = []
    model = current_app.larigira.controller.monitor.model
    ret = model.update_alarm(int(alarmid),
                             new_fields={'actions': [int(a) for a in
                                                     new_actions]})
    return jsonify(dict(updated=alarmid, ret=ret))



import logging
log = logging.getLogger('mpdrandom')
import random

from mpd import MPDClient

from .config import get_conf


def generate_by_artist(spec):
    '''choose HOWMANY random artists, and for each one choose a random song'''
    spec.setdefault('howmany', 1)
    log.info('generating')
    conf = get_conf()
    c = MPDClient(use_unicode=True)
    c.connect(conf['MPD_HOST'], conf['MPD_PORT'])

    artists = c.list('artist')
    log.debug("got %d artists" % len(artists))
    if not artists:
        raise ValueError("no artists in your mpd database")
    for _ in range(spec['howmany']):
        artist = random.choice(artists)
        yield random.choice(c.find('artist', artist))['file']



'''
script audiogenerator: uses an external program to generate audio URIs

a script can be any valid executable in
$XDG_CONFIG_DIR/larigira/scripts/<name>; for security reasons, it must be
executable and owned by the current user. The audiospec can specify arguments
to the script, while the environment cannot be customized (again, this is for
security reasons).

The script should assume a minimal environment, and being run from /.  It must
output one URI per line; please remember that URI must be understood from mpd,
so file paths are not valid; file:///file/path.ogg is a valid URI instead.
The output MUST be UTF-8-encoded.
Empty lines will be skipped.  stderr will be logged, so please be careful.  any
non-zero exit code will result in no files being added.and an exception being
logged.
'''
import logging
import os
import subprocess

from .config import get_conf
log = logging.getLogger('audioscript')


def generate(spec):
    '''
    Recognized arguments (fields in spec):
        - name [mandatory]            script name
        - args [default=empty]   arguments, space-separated
    '''
    conf = get_conf()
    spec.setdefault('args', '')
    args = spec['args'].split()
    for attr in ('name', ):
        if attr not in spec:
            raise ValueError("Malformed audiospec: missing '%s'" % attr)

    if '/' in spec['name']:
        raise ValueError("Script name is a filename, not a path ({} provided)"
                         .format(spec['name']))
    scriptpath = os.path.join(conf['SCRIPTS_PATH'], spec['name'])
    if not os.path.exists(scriptpath):
        raise ValueError("Script %s not found", spec['name'])
    if not os.access(scriptpath, os.R_OK | os.X_OK):
        raise ValueError("Insufficient privileges for script %s" % scriptpath)
    if os.stat(scriptpath).st_uid != os.getuid():
        raise ValueError("Script %s owned by %d, should be owned by %d"
                         % (spec['name'], os.stat(scriptpath).st_uid,
                            os.getuid()))
    try:
        log.info('Going to run {}'.format([scriptpath] + args))
        env = dict(
            HOME=os.environ['HOME'],
            PATH=os.environ['PATH'],
            MPD_HOST=conf['MPD_HOST'],
            MPD_PORT=str(conf['MPD_PORT'])
        )
        if 'TMPDIR' in os.environ:
            env['TMPDIR'] = os.environ['TMPDIR']
        out = subprocess.check_output([scriptpath] + args,
                                      env=env,
                                      cwd='/')
    except subprocess.CalledProcessError as exc:
        log.error("Error %d when running script %s" %
                  (exc.returncode, spec['name']))
        return []

    out = out.decode('utf-8')
    out = [p for p in out.split('\n') if p]
    logging.debug('Script %s produced %d files' % (spec['name'], len(out)))
    return out
generate.description = 'Generate audio through an external script. Experts only.'



import sys
import os

from setuptools import setup
from setuptools.command.test import test as TestCommand


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as buf:
        return buf.read()


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)

setup(name='larigira',
      version='0.3',
      description='A radio automation based on MPD',
      long_description=read('README.rst'),
      author='boyska',
      author_email='piuttosto@logorroici.org',
      license='AGPL',
      packages=['larigira', 'larigira.dbadmin'],
      install_requires=[
          'pyxdg',
          'gevent',
          'flask-bootstrap',
          'python-mpd2',
          'wtforms',
          'Flask-WTF',
          'flask==0.11',
          'pytimeparse',
          'tinydb'
      ],
      tests_require=['pytest', 'pytest-timeout'],
      cmdclass={'test': PyTest},
      zip_safe=False,
      include_package_data=True,
      entry_points={
          'console_scripts': ['larigira=larigira.larigira:main',
                              'larigira-timegen=larigira.timegen:main',
                              'larigira-audiogen=larigira.audiogen:main',
                              'larigira-dbmanage=larigira.event_manage:main'],
          'larigira.audiogenerators': [
              'mpd = larigira.audiogen_mpdrandom:generate_by_artist',
              'static = larigira.audiogen_static:generate',
              'randomdir = larigira.audiogen_randomdir:generate',
              'script = larigira.audiogen_script:generate',
          ],
          'larigira.timegenerators': [
              'frequency = larigira.timegen_every:FrequencyAlarm',
              'single = larigira.timegen_every:SingleAlarm',
          ],
          'larigira.timeform_create': [
              'single = larigira.timeform_single:SingleAlarmForm',
              'frequency = larigira.timeform_frequency:FrequencyAlarmForm',
          ],
          'larigira.timeform_receive': [
              'single = larigira.timeform_single:singlealarm_receive',
              'frequency = larigira.timeform_frequency:frequencyalarm_receive',
          ],
          'larigira.audioform_create': [
              'static = larigira.audioform_static:StaticAudioForm',
              'script = larigira.audioform_script:ScriptAudioForm',
              'randomdir = larigira.audioform_randomdir:Form',
          ],
          'larigira.audioform_receive': [
              'static = larigira.audioform_static:staticaudio_receive',
              'script = larigira.audioform_script:scriptaudio_receive',
              'randomdir = larigira.audioform_randomdir:receive',
          ],
      },
      classifiers=[
          "License :: OSI Approved :: GNU Affero General Public License v3",
          "Programming Language :: Python :: 3",
      ]
      )



from __future__ import print_function
import logging
import gc
from copy import deepcopy

from greenlet import greenlet
from flask import current_app, Blueprint, Flask, jsonify, render_template
from flask_bootstrap import Bootstrap

from .dbadmin import db
from .config import get_conf

rpc = Blueprint('rpc', __name__, url_prefix='/api')
viewui = Blueprint('view', __name__, url_prefix='/view',
                   template_folder='templates')


def send_to_parent(kind, *args):
    '''similar to the behaviour of a ParentedLet'''
    if not hasattr(current_app, 'queue'):
        logging.debug('no parent queue; aborting send')
        return
    msg = {
        'emitter': current_app._get_current_object(),
        'class': current_app._get_current_object().__class__.__name__,
        'kind': kind,
        'args': args
    }
    current_app.queue.put(msg)


@rpc.route('/')
def rpc_index():
    rules = list(current_app.url_map.iter_rules())
    return jsonify({'rules': [r.rule for r in rules
                              if r.rule.startswith('/api')]})


@rpc.route('/refresh')
def rpc_refresh():
    send_to_parent('refresh')
    return jsonify(dict(status='ok'))


def get_scheduled_audiogen():
    larigira = current_app.larigira
    running = larigira.controller.monitor.running
    events = {t: {} for t in running.keys()}
    for timespec_eid in events:
        orig_info = running[timespec_eid]
        info = events[timespec_eid]
        info['running_time'] = orig_info['running_time'].isoformat()
        info['audiospecs'] = orig_info['audiospecs']
        info['timespec'] = orig_info['timespec']
        info['timespec']['actions'] = {aid: spec
                                       for aid, spec
                                       in zip(info['timespec']['actions'],
                                              info['audiospecs'])
                                       }
        info['greenlet'] = hex(id(orig_info['greenlet']))
    return events


@viewui.route('/status/running')
def ui_wip():
    audiogens = get_scheduled_audiogen()
    return render_template('running.html',
                           audiogens=sorted(
                               audiogens.items(),
                               key=lambda x: x[1]['running_time'])
                           )


@rpc.route('/debug/running')
def rpc_wip():
    def treeify(flat):
        roots = [obid for obid in flat if flat[obid]['parent'] not in flat]
        tree = deepcopy(flat)
        for obid in tree:
            tree[obid]['children'] = {}

        to_remove = []
        for obid in tree:
            if obid in roots:
                continue
            if obid not in tree:
                current_app.logger.warn('How strange, {} not in tree'
                                        .format(obid))
                continue
            tree[tree[obid]['parent']]['children'][obid] = tree[obid]
            to_remove.append(obid)

        for obid in to_remove:
            del tree[obid]
        return tree

    greenlets = {}
    for ob in filter(lambda obj: isinstance(obj, greenlet),
                     gc.get_objects()):
        objrepr = {
            'repr': repr(ob),
            'class': ob.__class__.__name__,
        }
        if hasattr(ob, 'parent_greenlet') and ob.parent_greenlet is not None:
            objrepr['parent'] = hex(id(ob.parent_greenlet))
        else:
            objrepr['parent'] = hex(id(ob.parent)) \
                    if ob.parent is not None else None
        if hasattr(ob, 'doc'):
            objrepr['doc'] = ob.doc.split('\n')[0]
        elif ob.__doc__:
            objrepr['doc'] = ob.__doc__.split('\n')[0]

        greenlets[hex(id(ob))] = objrepr

    # TODO: make it a tree

    return jsonify(dict(greenlets=greenlets,
                        greenlets_tree=treeify(greenlets),
                        audiogens=get_scheduled_audiogen(),
                        ))


def create_app(queue, larigira):
    app = Flask('larigira')
    app.config.update(get_conf())
    Bootstrap(app)
    app.register_blueprint(rpc)
    app.register_blueprint(viewui)
    app.register_blueprint(db)
    app.queue = queue
    app.larigira = larigira
    return app



from __future__ import print_function
import logging
import signal

import gevent
from gevent.queue import Queue
from mpd import MPDClient, ConnectionError, CommandError

from .event import Monitor
from .eventutils import ParentedLet, Timer
from .audiogen import audiogenerate
from .unused import UnusedCleaner


def get_mpd_client(conf):
    client = MPDClient(use_unicode=True)
    client.connect(conf['MPD_HOST'], conf['MPD_PORT'])

    return client


class MpcWatcher(ParentedLet):
    def __init__(self, queue, conf, client=None):
        ParentedLet.__init__(self, queue)
        self.log = logging.getLogger(self.__class__.__name__)
        self.conf = conf
        self.client = client  # assume it is already connected, or None

    def refresh_client(self):
        self.client = get_mpd_client(self.conf)

    def do_business(self):
        while True:
            try:
                if self.client is None:
                    self.refresh_client()

                self.log.info('idling')
                status = self.client.idle()[0]
            except ConnectionError:
                # TODO: should we emit an error just in case?
                self.client = None
                continue
            self.log.info(status)
            yield ('mpc', status)


class Player:
    def __init__(self, conf):
        self.conf = conf
        self.log = logging.getLogger(self.__class__.__name__)
        self.min_playlist_length = 10

    def _get_mpd(self):
        mpd_client = MPDClient(use_unicode=True)
        mpd_client.connect(self.conf['MPD_HOST'], self.conf['MPD_PORT'])
        return mpd_client

    def check_playlist(self):
        mpd_client = self._get_mpd()
        songs = mpd_client.playlist()
        current = mpd_client.currentsong()
        pos = int(current.get('pos', 0)) + 1
        if(len(songs) - pos >= self.min_playlist_length):
            return
        self.log.info('need to add new songs')
        picker = gevent.Greenlet(audiogenerate,
                                 self.conf['CONTINOUS_AUDIODESC'])

        def add(greenlet):
            uris = greenlet.value
            for uri in uris:
                assert type(uri) is str, type(uri)
                mpd_client.add(uri.strip())
        picker.link_value(add)
        picker.start()

    def enqueue(self, songs):
        mpd_client = self._get_mpd()
        assert type(songs) is dict
        assert 'uris' in songs
        spec = [aspec.get('nick', aspec.eid) for aspec in songs['audiospecs']]
        for uri in reversed(songs['uris']):
            assert type(uri) is str
            self.log.info('Adding {} to playlist (from <{}>:{}={})'.
                          format(uri, songs['timespec'].get('nick', ''),
                                 songs['aids'], spec))
            insert_pos = 0 if len(mpd_client.playlistid()) == 0 else \
                int(mpd_client.currentsong().get('pos', 0)) + 1
            try:
                mpd_client.addid(uri, insert_pos)
            except CommandError:
                self.log.exception("Cannot insert song {}".format(uri))


class Controller(gevent.Greenlet):
    def __init__(self, conf):
        gevent.Greenlet.__init__(self)
        self.log = logging.getLogger(self.__class__.__name__)
        self.conf = conf
        self.q = Queue()
        self.player = Player(self.conf)
        self.tmpcleaner = UnusedCleaner(conf)
        if 'DB_URI' in self.conf:
            self.monitor = Monitor(self.q, self.conf)
            self.monitor.parent_greenlet = self
        else:
            self.monitor = None

    def _run(self):
        if self.monitor is not None:
            self.monitor.start()
        mw = MpcWatcher(self.q, self.conf, client=None)
        mw.parent_greenlet = self
        mw.start()
        t = Timer(int(self.conf['CHECK_SECS']) * 1000, self.q)
        t.parent_greenlet = self
        t.start()
        # at the very start, run a check!
        gevent.Greenlet.spawn(self.player.check_playlist)
        while True:
            value = self.q.get()
            self.log.debug('<- %s' % str(value))
            # emitter = value['emitter']
            kind = value['kind']
            args = value['args']
            if kind == 'timer' or (kind == 'mpc' and
                                   args[0] in ('player', 'playlist')):
                gevent.Greenlet.spawn(self.player.check_playlist)
                self.tmpcleaner.check_playlist()
            elif kind == 'mpc':
                pass
            elif kind == 'uris_enqueue':
                try:
                    self.player.enqueue(args[0])
                except AssertionError:
                    raise
                except Exception:
                    self.log.exception("Error while adding to queue; "
                                       "bad audiogen output?")
                else:
                    for fname in args[0]['uris']:
                        self.tmpcleaner.watch(fname)
            elif (kind == 'signal' and args[0] == signal.SIGALRM) or \
                    kind == 'refresh':
                # it's a tick!
                self.log.debug("Reload")
                self.monitor.q.put(dict(kind='forcetick'))
                gevent.Greenlet.spawn(self.player.check_playlist)
            else:
                self.log.warning("Unknown message: %s" % str(value))



import os
import logging
import shutil
import random
from tempfile import mkstemp
import fnmatch


def scan_dir(dirname, extension=None):
    if extension is None:
        fnfilter = lambda fnames: fnames
    else:
        fnfilter = lambda fnames: fnmatch.filter(fnames, extension)
    for root, dirnames, filenames in os.walk(dirname):
        for fname in fnfilter(filenames):
            yield os.path.join(root, fname)


def generate(spec):
    '''
    resolves audiospec-randomdir

    Recognized arguments:
        - paths [mandatory]    list of source paths
        - howmany [default=1]  number of audio files to pick
    '''
    spec.setdefault('howmany', 1)
    for attr in ('paths', ):
        if attr not in spec:
            raise ValueError("Malformed audiospec: missing '%s'" % attr)

    found_files = set()
    for path in spec['paths']:
        if not os.path.exists(path):
            logging.warn("Can't find requested path: %s" % path)
            continue
        if os.path.isfile(path):
            found_files.add(path)
        elif os.path.isdir(path):
            found_files.update(scan_dir(path))

    picked = random.sample(found_files, int(spec['howmany']))

    for path in picked:
        tmp = mkstemp(suffix=os.path.splitext(path)[-1],
                      prefix='audiogen-randomdir-')
        os.close(tmp[0])
        shutil.copy(path, tmp[1])
        yield 'file://{}'.format(tmp[1])

generate.description = 'Picks random files from a specified directory'
